-------------------------------------------------------------------------------
--! @file       CipherCore_Control.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Control unit for JAMBU-AES
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore_Control is
    generic (
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! Input
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;

        mode            : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(4                -1 downto 0);
        is_data         : in  std_logic;

        --! Control
        en_key          : out std_logic;
        en_init         : out std_logic;
        sel_xorx        : out std_logic_vector(2             -1 downto 0);
        sel_xory        : out std_logic_vector(2             -1 downto 0);
        sel_tag         : out std_logic;
        aes_en          : out std_logic;
        aes_almost_done : in  std_logic;         

        --! Output
        bdo_ready       : in  std_logic;
        bdo_valid       : out std_logic;
        msg_auth_done   : out std_logic
    );
end entity CipherCore_Control;

architecture behavior of CipherCore_Control is
    
    type state_type is (
        S_START,            S_KEY_CHECK,
        S_LDKEY_1,          S_LDKEY_2,
        S_LDKEY_3,          S_LDKEY_4,
        S_WAIT_NPUB,        S_INIT,
        S_AD,               S_DATA,
        S_TAG,              S_OUT,
        S_WAIT_DONE);
    signal state, nstate      : state_type;
    signal nnstate, nnstate_r : state_type;


    signal bdi_rdy        : std_logic;
    signal bdo_vld        : std_logic;
    signal is_decrypt     : std_logic;
begin
    bdi_ready <= bdi_rdy;
    bdo_valid <= bdo_vld;

    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                state <= S_START;
                nnstate_r <= S_INIT;
            else
                state <= nstate;
                nnstate_r <= nnstate;
                
                if state = S_WAIT_NPUB then
                    is_decrypt <= mode;
                end if;

            end if;
        end if;
    end process;

    p_state:
    process( state, bdi_valid, is_data, bdi_eot, key_valid, key_update, bdo_ready, is_decrypt, aes_almost_done)
    begin
        --! Default values
        nstate          <= state;
        nnstate         <= nnstate_r;
        bdi_rdy         <= '0';
        bdo_vld         <= '0';
        key_ready       <= '0';

        en_key          <= '0';       
        en_init         <= '0';
        sel_xorx        <= "00";
        sel_xory        <= "00";
        sel_tag         <= '0';
        aes_en          <= '0';
        msg_auth_done   <= '0';

        case state is
            when S_START =>
                nstate       <= S_KEY_CHECK;

            when S_KEY_CHECK =>
                if (key_update = '0' and bdi_valid = '1') then
                    nstate      <= S_WAIT_NPUB;
                elsif (key_update = '1' and key_valid = '1') then
                    nstate      <= S_LDKEY_1;
                end if;

            when S_LDKEY_1 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_LDKEY_2;
                    en_key <= '1';
                end if;

            when S_LDKEY_2 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_LDKEY_3;
                    en_key <= '1';
                end if;

            when S_LDKEY_3 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_LDKEY_4;
                    en_key <= '1';
                end if;

            when S_LDKEY_4 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_WAIT_NPUB;
                    en_key <= '1';
                end if;

            when S_WAIT_NPUB =>
                en_init <= '1';
                nnstate <= S_INIT;
                bdi_rdy <= '1';
                if (bdi_valid = '1') then                    
                    aes_en <= '1';
                    nstate <= S_WAIT_DONE;                   
                end if;

            when S_INIT =>
                nnstate <= S_AD;
                aes_en   <= '1';
                sel_xorx <= "10";
                sel_xory <= "10";                                
                nstate   <= S_WAIT_DONE;
                
            when S_AD =>                
                sel_xory <= "01";
                if (bdi_eot = '1') then
                    nnstate  <= S_DATA;
                end if;
                bdi_rdy <= '1';
                if (bdi_valid = '1') then                       
                    aes_en  <= '1';
                    nstate  <= S_WAIT_DONE;
                end if;

            when S_DATA =>
                if (bdi_eot = '1') then
                    nnstate  <= S_TAG;
                end if;                
                if (bdi_valid = '1' and is_data = '1' and bdo_ready = '1') 
                    or (bdi_valid = '1' and is_data = '0') 
                then
                    bdi_rdy <= '1';
                    if (is_data = '1') then
                        bdo_vld <= '1';
                    end if;
                    
                    if (is_decrypt = '1') then
                        sel_xorx <= "01";                        
                    end if;
                    aes_en  <= '1';
                    nstate  <= S_WAIT_DONE;
                end if;

            when S_TAG =>
                nnstate  <= S_OUT;
                sel_xorx <= "11";
                sel_xory <= "11";                                
                aes_en   <= '1';
                nstate   <= S_WAIT_DONE;
            
                
            when S_OUT =>
                sel_tag <= '1';
                if (is_decrypt = '0' and bdo_ready = '1') then
                    bdo_vld <= '1';
                    nstate  <= S_START;
                elsif (is_decrypt = '1' and bdi_valid = '1') then
                    bdi_rdy <= '1';
                    msg_auth_done <= '1';
                    nstate  <= S_START;
                end if;
                
            when S_WAIT_DONE =>
                if (aes_almost_done = '1') then
                    nstate <= nnstate_r;
                end if;

        end case;
    end process;

end behavior;

