-------------------------------------------------------------------------------
--! @file       CipherCore_Datapath.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Datapath unit for JAMBU-AES
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;
use work.AES_pkg.all;

entity CipherCore_Datapath is
    generic (
        G_SBOX_LOGIC        : boolean := False;
        G_MAX_LEN           : integer := SINGLE_PASS_MAX
    );
    port    (
        clk                 : in  std_logic;
        rst                 : in  std_logic;

        --! Input
        bdi                 : in  std_logic_vector(64            -1 downto 0);
        bdi_valid_bytes     : in  std_logic_vector(64/8          -1 downto 0);
        bdi_pad_loc         : in  std_logic_vector(64/8          -1 downto 0);
        key                 : in  std_logic_vector(32            -1 downto 0);

        --! Control
        en_key              : in  std_logic;
        en_init             : in  std_logic;
        sel_xorx            : in  std_logic_vector(2             -1 downto 0);
        sel_xory            : in  std_logic_vector(2             -1 downto 0);
        sel_tag             : in  std_logic;
        aes_en              : in  std_logic;
        aes_almost_done     : out std_logic;

        --! Output
        bdo                 : out std_logic_vector(64            -1 downto 0);
        msg_auth_valid      : out std_logic
    );
end entity CipherCore_Datapath;

architecture structure of CipherCore_Datapath is
    constant ZEROS          : std_logic_vector(64               -1 downto 0)
                            := (others => '0');

    --! Registers
    signal reg_key          : std_logic_vector(128              -1 downto 0);
    signal sr               : std_logic_vector(64               -1 downto 0);
    --! Signals
    signal ss               : std_logic_vector(128              -1 downto 0);
    signal aeso             : std_logic_vector(128              -1 downto 0);
    signal xx               : std_logic_vector(64               -1 downto 0);
    signal yy               : std_logic_vector(64               -1 downto 0);
    signal uu               : std_logic_vector(64               -1 downto 0);
    signal vv               : std_logic_vector(64               -1 downto 0);
    signal pt               : std_logic_vector(64               -1 downto 0);
    signal to_xor_x         : std_logic_vector(64               -1 downto 0);
    signal to_xor_y         : std_logic_vector(3                -1 downto 0);
    signal output           : std_logic_vector(64               -1 downto 0);
    signal tag_data         : std_logic_vector(64               -1 downto 0);
begin
    pRegs:
    process(clk)
    begin
        if rising_edge(clk) then
            if (en_key = '1') then
                reg_key <= reg_key(95 downto 0) & key;
            end if;
            if (en_init = '1') then
                sr <= (others => '0');
            elsif (aes_en = '1') then
                sr <= sr xor uu;
            end if;
        end if;
    end process;

    with sel_xory select
    to_xor_y <= "000" when "00",
                "001" when "01",
                "101" when "10",
                "011" when others;

    with sel_xorx select
    to_xor_x <= bdi when "00",
                pt       when "01",      --! Decryption
                (others => '0') when others;

    gPt:
    for i in 64/8-1 downto 0 generate

        pt(i*8+7 downto i*8) <= output(i*8+7 downto i*8)
                                when bdi_valid_bytes(i) = '1'
                                else (bdi_pad_loc(i) & "0000000");
    end generate;

    xx <= aeso(63 downto 0);
    yy <= aeso(127 downto 64);
    uu <= xx xor to_xor_x;
    vv(63 downto 59) <= yy(63 downto 59) xor sr(63 downto 59);
    vv(58 downto 56) <= yy(58 downto 56) xor sr(58 downto 56) xor to_xor_y;
    vv(55 downto  0) <= yy(55 downto  0) xor sr(55 downto  0);
    output <= vv xor bdi;

    ss <= bdi & ZEROS(63 downto 0) when en_init = '1' else vv & uu;

    u_aes: entity work.AES_Enc_KOF(structure)
    generic map (G_SBOX_LOGIC => G_SBOX_LOGIC)
    port map (clk => clk, rst => rst,
              din       => ss,
              key       => reg_key,
              dout      => aeso,
              start     => aes_en,
              almost_done => aes_almost_done);

    bdo <= tag_data when sel_tag = '1' else output;
    tag_data <= xx xor yy xor sr;
    msg_auth_valid <= '1' when tag_data = bdi else '0';
end architecture structure;